#include <avr/sleep.h> 
#include <SoftwareSerial.h>
#include <SPI.h>
#include <DS3232RTC.h>
#include <SD.h>

//SD card module
File log_file;
String file_name = "dendro.csv"; //8.3 filename convention only
const int chipselect_pin = 10;
const int sd_en_pin = A1;

//DS3231 RTC module
DS3232RTC RTC;
const int wakeup_pin = 2; //RTC interupt pin
const int sleep_interval = 1; //minutes, sleep interval

//HM-10 bluetooth module
SoftwareSerial BTserial(8, 7); // RX | TX
const int blue_state_pin = A2;
const int blue_en_pin = A3;

//tread depth gauge
const int clock_pin = 3;
const int data_pin = 4;
const int units_pin = 5;
const int tdg_en_pin = 6;

boolean data[24] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1}; //24 bit boolean array for storing the received data
boolean mm_in = 0; //0 for mm, 1 for inch
int received_bit_location = 0;
long last_received = 0;
float gauge_value = 0.0;

//battery voltage monitoring
const int battery_pin = A0;
const float voltage_ref = 3.297; //V, VCC reference voltage

void setup() {
  Serial.begin(9600);

  //DS3231 RTC setup
  pinMode(wakeup_pin,INPUT_PULLUP);
  RTC.setAlarm(DS3232RTC::ALM1_MATCH_DATE, 0, 0, 0, 1);
  RTC.setAlarm(DS3232RTC::ALM2_MATCH_DATE, 0, 0, 0, 1);
  RTC.alarm(DS3232RTC::ALARM_1);
  RTC.alarm(DS3232RTC::ALARM_2);
  RTC.alarmInterrupt(DS3232RTC::ALARM_1, false);
  RTC.alarmInterrupt(DS3232RTC::ALARM_2, false);
  RTC.squareWave(DS3232RTC::SQWAVE_NONE);
  
//  Uncomment the block to set the time on your RTC. Remember to comment it again 
//  otherwise you will set the time at everytime you upload the sketch
//  tmElements_t tm;
//  tm.Hour = 22;               // set the RTC time
//  tm.Minute = 35;
//  tm.Second = 00;
//  tm.Day = 18;
//  tm.Month = 10;
//  tm.Year = 2022 - 1970;      // tmElements_t.Year is the offset from 1970
//  RTC.write(tm); 
            
  time_t t;
  t = RTC.get();
  if ((minute(t)+sleep_interval) > 59){
    RTC.setAlarm(DS3232RTC::ALM1_MATCH_MINUTES , 0, minute(t)+sleep_interval-60, 0, 0);
  }
  else{
    RTC.setAlarm(DS3232RTC::ALM1_MATCH_MINUTES , 0, minute(t)+sleep_interval, 0, 0);
  }
  RTC.alarm(DS3232RTC::ALARM_1);
  RTC.squareWave(DS3232RTC::SQWAVE_NONE);
  RTC.alarmInterrupt(DS3232RTC::ALARM_1, true);

  //SD card setup
  pinMode(sd_en_pin, OUTPUT);
  digitalWrite(sd_en_pin, HIGH); //connect power to SD card module
  
  if (!SD.begin(chipselect_pin)) {
      Serial.println(F("initialization failed!"));
      return;
  }
  Serial.println(F("initialization done."));
  
  //add column header to CSV file
  log_file = SD.open(file_name, FILE_WRITE);
  if (log_file){
    log_file.println(" , , , , , , ,");
    log_file.println("Year,Month,Day,Hour,Minute,Dendrometer (mm),Temperature (°C),Battery Voltage (V)");
  }
  log_file.close();
  digitalWrite(sd_en_pin, LOW); //disconnect power to SD card module
  
  //bluetooth setup
  BTserial.begin(19200);
  pinMode(blue_state_pin, INPUT);
  pinMode(blue_en_pin, OUTPUT);
  digitalWrite(blue_en_pin, HIGH);
  
  //tire depth gauge setup
  pinMode(clock_pin,INPUT);
  pinMode(data_pin,INPUT);
  pinMode(units_pin,OUTPUT);
  pinMode(tdg_en_pin,OUTPUT);
  digitalWrite(units_pin,HIGH);
  digitalWrite(tdg_en_pin,HIGH);
}

void loop() {

  //INCLUDE CONDITIONS TO DEACTIVATE BLUETOOTH HERE
  
  while (digitalRead(blue_state_pin) == HIGH){ //bluetooth is connected
    if (BTserial.available()){ //wait for data received
      byte command = BTserial.read();
      if(command == 'H'){  
        BTserial.println(F("B -> Battery Voltage, D -> Data Transfer"));
      }
      else if (command == 'B'){
        BTserial.print(F("Battery Voltage: "));
        BTserial.print(analogRead(battery_pin)*(voltage_ref/1023)*2);
      }
      else if(command=='D'){
        digitalWrite(sd_en_pin, HIGH); //connect power to SD card module
        File dataFile = SD.open("dendro.CSV");
        if (dataFile)
        {
          while (dataFile.available())
          {
             BTserial.write(dataFile.read());
          }
          dataFile.close();
        }
        digitalWrite(sd_en_pin, LOW); //disconnect power to SD card module
      }
    }
    delay(100);
  }
  
  BTserial.println("AT+SLEEP"); //put bluetooth module to sleep

  float battery_voltage = analogRead(battery_pin)*(voltage_ref/1023)*2;
  delay(3000); //delay to get latest digital signal from digital gauge

  //TODO: only write to SD card when sleep period has expired, not right after Bluetooth has been connected
  //write data to SD card
  digitalWrite(sd_en_pin, HIGH); //connect power to SD card module
  Serial.print(F("Initializing SD card..."));
  if (!SD.begin(chipselect_pin)) {
      Serial.println(F("initialization failed!"));
      return;
  }
  Serial.println(F("initialization done."));
  write_data(gauge_value, battery_voltage); //write data to SD card
  digitalWrite(sd_en_pin, LOW); //disconnect power to SD card module

  time_t t;
  t = RTC.get();
  if ((minute(t)+sleep_interval) > 59){
    RTC.setAlarm(DS3232RTC::ALM1_MATCH_MINUTES , 0, minute(t)+sleep_interval-60, 0, 0);
  }
  else{
    RTC.setAlarm(DS3232RTC::ALM1_MATCH_MINUTES , 0, minute(t)+sleep_interval, 0, 0);
  }
  RTC.alarm(DS3232RTC::ALARM_1);
  
  //switch from mm to in and back to mm; minimum every four minutes, to keep digital gauge awake
  for (int i = 0; i <= 1; i++){
    digitalWrite(units_pin, LOW);
    delay(200);
    digitalWrite(units_pin, HIGH);
    delay(200);
  }

  //delay(1000); //needed if serial.print directly before go_to_sleep()
  go_to_sleep();
}

void go_to_sleep(){
  sleep_enable();
  detachInterrupt(digitalPinToInterrupt(clock_pin));
  attachInterrupt(digitalPinToInterrupt(wakeup_pin), wakeup_ISR, LOW);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_cpu(); 
  // continues after sleep_cpu() once wakeup_ISR is finished
}

//wake up interrupt service routine
void wakeup_ISR(){ 
  sleep_disable();
  detachInterrupt(digitalPinToInterrupt(wakeup_pin));
  attachInterrupt(digitalPinToInterrupt(clock_pin), clock_ISR, RISING); 
}

void write_data(float gauge_val, float battery_voltage){
  time_t p;
  p = RTC.get();
  float T_celsius = RTC.temperature() / 4.0;
  log_file = SD.open(file_name, FILE_WRITE);

  if (log_file) {
    log_file.println(String(year(p))+","
                     +String(month(p))+","
                     +String(day(p))+","
                     +String(hour(p))+","
                     +String(minute(p))+","
                     +String(gauge_val)+","
                     +String(T_celsius,2)+","
                     +String(battery_voltage));
    log_file.close();
  } else {
    Serial.println("error opening "+ file_name);
  }
  
//  //opening file for reading and writing content to serial monitor
//  log_file = SD.open(file_name);
//  if (log_file) {
//    Serial.println(file_name);
//    while (log_file.available()) {
//      Serial.write(log_file.read());
//    }
//    log_file.close();
//  } else {
//    Serial.println("error opening "+ file_name);
//  }
 }

//digital gauge process data
void process_data(){
  
  gauge_value = 0.0;
  
  if(data[23] == 0){ //if it's in the mm mode
    mm_in = 0;
    //converting binary to decimal value
    for (int i = 0; i<=15; i++){
    gauge_value += data[i]*pow(2,i)/100.0;
    }
      if (data[20] == 1){
      gauge_value = gauge_value * -1; //add the negative sign if it exists
      }
  }
  
  if(data[23] == 1){ //if it's in the inch mode
    mm_in = 1;
    //converting binary to decimal value
    for (int i = 1; i<=16; i++){
    gauge_value += data[i]*pow(2,(i-1))/1000.0;
    }
      if (data[0]  == 1){
      gauge_value += 0.0005; //add the 0.5 mil sign if it exists  
      }
      if (data[20] == 1){
      gauge_value = gauge_value * -1; //add the negative sign if it exists
      }
  }
  //Serial.println(gauge_value); //for troubleshooting
}

//digital gauge clock interrupt service routine
void clock_ISR(){
  if (millis() - last_received > 2){
    received_bit_location = 0;
  }
  
  if (received_bit_location <= 23){
    data[received_bit_location] = digitalRead(data_pin); 
    received_bit_location++;
  } 
  
  if (received_bit_location == 24){
    received_bit_location = 0;
    process_data();
  }
  
  last_received = millis();
}
